from papercandy.test import *
from papercandy.train import *
from papercandy.config import *
from papercandy.drawing import *
from papercandy.network import *
from papercandy.version import *
from papercandy.dataloader import *


__repo__ = "https://github.com/ATATC/PaperCandy"
"""
NOTICE: This version is built with PyTorch. Consider reimplementing the abstract classes in the `core` package to
adapt to the framework that you use if you are not a PyTorch user.

Visit the repository for details.
"""
